#!/usr/bin/env python
"""
Laser dynamics simulation

Authors: Mohammad Mayy and Alexander V. Gavrilenko

"""

import sys,math
import numpy

# Import rappture libraries
try:
	import Rappture

# Import fail!
except ImportError:
	
	# If rappture not present, print error in python3000 format
	try:
		print("Running {a} without rappture!".format(a=sys.argv[0]))
	
	# Older version of python error
	except AttributeError:
		print "Running %s without rappture!" %(sys.argv[0])

# Inclusive range function
def inclusive_range(min, max, step=1):
	return (range(min, max+step, step))

def outputLog(driver, stimEmissionList, step, maxPopInv, pumpingPowerValue):
	# Initialize variables
	totalStimEmitPower = 0
	
	# Riemann Sum of stimEmission to get total stimulated emission power
	for point in range(0, len(stimEmissionList)-1, 1):
		totalStimEmitPower += (stimEmissionList[point]+stimEmissionList[point+1])/2*step
		
	outputLogText="""
===========================
Random laser dynamics
===========================
Pumping energy density:\t\t%f J/cm^2
Stim. emission energy density:\t%f J/cm^2
Stim. emission quantum yield:\t%3.2f%%
Max population inversion:\t%g 1/cm^3\n""" %(pumpingPowerValue, totalStimEmitPower, totalStimEmitPower/pumpingPowerValue*100, maxPopInv)
	driver.put('output.log', outputLogText)

def addPlotValue(driver, curveId, xVal, yVal):
	plotPath = 'output.curve('+curveId+').component.xy'
	driver.put(plotPath, '%g %g\n' % (xVal, yVal), append=1)

def labelXYGraph(driver,groupName,graphPath,graphName,xAxisName,xAxisUnits,yAxisName,yAxisUnits):
	driver.put(graphPath+'.about.group',groupName, append=0)
	driver.put(graphPath+'.about.label',graphName, append=0)
	driver.put(graphPath+'.xaxis.label',xAxisName, append=0)
	driver.put(graphPath+'.xaxis.units',xAxisUnits, append=0)
	driver.put(graphPath+'.yaxis.label',yAxisName,append=0)
	driver.put(graphPath+'.yaxis.units',yAxisUnits, append=0)

def labelPulseShapePlot(driver,curveId):
	groupName  = 'Random laser dynamics simulation'
	graphPath  = 'output.curve('+curveId+')'
	graphName  = 'Pulse Shape'
	xAxisName  = 'Time'
	xAxisUnits = 'ns'
	yAxisName  = 'Power density'
	yAxisUnits = 'W/cm^2'
	labelXYGraph (driver,groupName,graphPath,graphName,xAxisName,xAxisUnits,yAxisName,yAxisUnits )

def labelStimEmissionPlot(driver,curveId):
	groupName  = 'Random laser dynamics simulation'
	graphPath  = 'output.curve('+curveId+')'
	graphName  = 'Stimulated Emission'
	xAxisName  = 'Time'
	xAxisUnits = 'ns'
	yAxisName  = 'Power density'
	yAxisUnits = 'W/cm^2'
	labelXYGraph (driver,groupName,graphPath,graphName,xAxisName,xAxisUnits,yAxisName,yAxisUnits )

def labelPopInvPlot(driver,curveId):
	groupName  = 'Random laser dynamics simulation'
	graphPath  = 'output.curve('+curveId+')'
	graphName  = 'Population Inversion'
	xAxisName  = 'Time'
	xAxisUnits = 'ns'
	yAxisName  = 'Population Inversion (cm-3)'
	yAxisUnits = ''  #cm^{-3}
	labelXYGraph (driver,groupName,graphPath,graphName,xAxisName,xAxisUnits,yAxisName,yAxisUnits )	

def plot2D(stimEmission, pulseShape, popInv):
	try:
		import pylab
	except ImportError:
		print ("MatPlotLib isn't properly installed!")	
		return
	
	stimEmissionMax = max(stimEmission)
	pulseShapeMax = max(pulseShape)
	popInvMax = max(popInv)
	
	for k in range(0 , len(stimEmission), 1):
		stimEmission[k] = stimEmission[k]/stimEmissionMax
		pulseShape[k] = pulseShape[k]/pulseShapeMax
		popInv[k] = popInv[k]/popInvMax
	
	pylab.ylim([0,2])
	pylab.plot(range(0, len(stimEmission), 1), stimEmission, range(0, len(pulseShape), 1), pulseShape, range(0, len(popInv), 1), popInv)
	pylab.show()

def main():
	
	# Constants
	numOfIterations = 1000
	calcTimeStep = 2e-12
	planckConst = 6.62606896e-34
	lightSpeed = 29979245800 # cm/s not m/s
	absCoeffSolid = 12.5
	fillingFactor = 0.0
	absCoeffMix = absCoeffSolid*(1-fillingFactor)
	emissionWvl = 1063.0
	stimEmissionCurveId = "stimEmission"
	popInvCurveId = "popInversion"
	pulseShapeCurveId = "pulseShape"
	
	#INPUTS
	#Read input from Rappture
	try:
		driver = Rappture.library(sys.argv[1])
		pumpingWvl = Rappture.Units.convert(driver.get("input.number(pumpingWvl).current"), "nm", units="off")
		pulseDuration = Rappture.Units.convert(driver.get("input.number(pulseDuration).current"), "s", units="off")
		spontEmitDecayTime = Rappture.Units.convert(driver.get("input.number(spontEmitDecayTime).current"), "s", units="off") #tao
		residenceTime = Rappture.Units.convert(driver.get("input.number(residenceTime).current"), "s", units="off")
		spontEmitCrossSect = float(driver.get("input.number(spontEmitCrossSect).current"))
		pumpingEnergy = float(driver.get("input.number(pumpingEnergy).current"))
		spontEmitQuantYield = float(driver.get("input.number(spontEmitQuantYield).current"))
		
	# Reading input values from Rappture fail! Set some values
	except NameError:
		pumpingWvl = 532.0 #nm
		pulseDuration = 5e-9
		spontEmitDecayTime = 1.6e-5
		residenceTime = 1e-11
		spontEmitCrossSect = 9e-20
		pumpingEnergy = 12.533141373155
		spontEmitQuantYield = 0.1356
		
		# Test values; If calculations are correct, the 3 curves will be on top of each other
		#spontEmitQuantYield = 1
		#spontEmitDecayTime = 1.6e-12
		#pumpingEnergy = 1
	
	# Convert wavelength to frequency
	pumpingFreq = lightSpeed/pumpingWvl/1e-7 #Hz
	emissionFreq = lightSpeed/emissionWvl/1e-7 #Hz
	
	# Calculate peak pumping power (value at the top of the gaussian pulse)
	peakPumpPwr = pumpingEnergy/(pulseDuration*math.sqrt(2*math.pi))  #1.0e9
	
	# Total time is eight pulse halfmaximum widths
	totalTime = pulseDuration*8
	
	# Pump for half of the time
	maxPumpTime = totalTime/2
	
	# Initialize the lists
	pulseShape = [0]
	stimEmission = [0]
	popInv = [0]
	
	# No population inversion nor stimulated emission before execution
	currentPopInv = 0.0
	currentEmission = 0.0
	
	#Split totalTime into printTimeStep chunks
	for currentTimeStep in numpy.arange(0, totalTime, totalTime/numOfIterations):
		
		#split printTimeStep into calcTimeStep
		for currentCalcTimeStep in numpy.arange(0, totalTime/numOfIterations, calcTimeStep):
			
			# Calculate current time
			currentTime = currentTimeStep+currentCalcTimeStep
			
			# Pumping power at the current calcTimeStep
			currentPumpPwr = peakPumpPwr*math.exp((-1)*(currentTime-maxPumpTime)**2.0/(2*pulseDuration**2.0))
			
			# Population inversion at the current calcTimeStep
			currentPopInv += (currentPumpPwr*absCoeffSolid/planckConst/pumpingFreq - \
				(currentPopInv/spontEmitDecayTime) - \
				((currentEmission*lightSpeed*spontEmitCrossSect*currentPopInv)/planckConst/emissionFreq))*calcTimeStep
				
			# Stimulated emission at the current calcTimeStep
			currentEmission += (-currentEmission/residenceTime + \
				(spontEmitQuantYield*(currentPopInv/spontEmitDecayTime)*planckConst*emissionFreq) + \
				(currentEmission*lightSpeed*spontEmitCrossSect*currentPopInv))*calcTimeStep
		
		# Add calculated values to the array
		Rappture.Utils.progress(currentTimeStep/totalTime*100,"Simulating")
		stimEmission.append(currentEmission/(residenceTime*absCoeffSolid))
		pulseShape.append(currentPumpPwr)
		popInv.append(currentPopInv)
	
	outputLog(driver, stimEmission, totalTime/numOfIterations, max(popInv), pumpingEnergy)
	
	# Try to plot using Rappture
	try:		
		# Label Rappture plots
		labelStimEmissionPlot(driver, stimEmissionCurveId)
		labelPopInvPlot(driver, popInvCurveId)
		labelPulseShapePlot(driver, pulseShapeCurveId)
		
		# CURRENTLY DOES NOT WORK, DISPLAYS A GRAPH WITH ONE POINT
		# Display total stimulated emission power density.
		#stimEmitPowerValue(driver, stimEmission, totalTime/numOfIterations)
		
		# Plot the xy values
		for currentTimeStep in numpy.arange(0, totalTime, totalTime/numOfIterations):
			addPlotValue(driver, pulseShapeCurveId, currentTimeStep, pulseShape.pop(0))			
			addPlotValue(driver, popInvCurveId, currentTimeStep, popInv.pop(0))
			addPlotValue(driver, stimEmissionCurveId, currentTimeStep, stimEmission.pop(0))
			
		# Notify rappture that script execution finished
		Rappture.result(driver)
	
	# Rappture plot fail!
	except UnboundLocalError:
		plot2D(stimEmission, pulseShape, popInv)
		

if __name__=="__main__":
	
	# Start main loop
	main()
